from django.conf import settings
from django.core.management.base import BaseCommand


class PartialRunserverCommand(BaseCommand):
    def handle(self, *args, **options):
        if not options["addrport"]:
            try:
                run_on = settings.RUNSERVER_ON
                if run_on:
                    options["addrport"] = run_on
            except AttributeError:
                pass
        return super().handle(*args, **options)



from importlib import import_module

from django.apps import AppConfig
import django.core.management

from runserveronhostname.management.commands._runserver import PartialRunserverCommand


def patched_load_command_class(app_name, name):
    """
    Adapted from `django.core.management.load_command_class`.

    Given a command name and an application name, return the Command
    class instance. Allow all errors raised by the import process
    (ImportError, AttributeError) to propagate.
    """
    module = import_module("%s.management.commands.%s" % (app_name, name))
    if name != 'runserver':
        return module.Command()

    Command = type('Command', (PartialRunserverCommand, module.Command), dict())
    return Command()


class RunserveronhostnameConfig(AppConfig):
    name = 'runserveronhostname'

    def ready(self):
        
        django.core.management.load_command_class = patched_load_command_class
        return super().ready()



from pathlib import Path
import platform

from django.conf import settings
from django.core.management.base import BaseCommand, CommandError

from runserveronhostname.hostfile_parser import Hostfile


class Command(BaseCommand):
    help = "Analyze your system hostfile to see if this project's RUNSERVER_ON hostname is mentioned."

    def add_arguments(self, parser):    # pragma: no cover - we don't need to test Django itself
        parser.add_argument(
            "--write",
            action="store_true",
            help="Print a hostfile with this project added",
        )
        parser.add_argument(
            "--status",
            action="store_true",
            help="Exit with 0 status if this project is already included, 1 if not",
        )
        parser.add_argument(
            "--file",
            action="store",
            help="Exit with 0 status if this project is already included, 1 if not",
        )

    def handle(self, *args, **options):
        try:
            runserver_on = settings.RUNSERVER_ON
        except AttributeError:
            self.stdout.write("RUNSERVER_ON not found in settings.")
            return
        
        target_hostname, _ = runserver_on.split(':', maxsplit=1)

        system = platform.system()
        if options['file']:
            hostfile = Path(options['file'])
        elif system in ('Darwin', 'Linux'):
            hostfile = Path('/etc/hosts')
        elif system in ('Windows',):
            raise CommandError("Windows has a known hostfile but we haven't dealt with it yet", returncode=2)
        else:
            raise CommandError(f'No implementation for {system}', returncode=2)

        with open(hostfile) as f:
            hosts = Hostfile(f)

        write_hostfile = options['write']

        if write_hostfile:
            self.stdout.write(str(hosts))
            if target_hostname not in hosts:
                self.stdout.write(f"127.0.0.1\t{target_hostname}")
        else:
            if target_hostname in hosts:
                self.stdout.write(f"{target_hostname} is already in {hostfile}.")
            else:
                err_msg = f"{target_hostname} not found in {hostfile}."
                if options['status']:
                    raise CommandError(err_msg, returncode=1)
                self.stderr.write(err_msg)



from collections import defaultdict
from collections.abc import Mapping
from typing import IO


class Hostfile(Mapping):
    def __init__(self, text_or_file: str|IO):
        if isinstance(text_or_file, str):
            self._contents = text_or_file.splitlines(keepends=True)
        else:
            self._contents = [l for l in text_or_file.readlines()]
        self._ips = defaultdict(list)
        self._hosts = defaultdict(list)

        for idx, line in enumerate(self._contents, start=1):
            line = line.strip()
            # skip comments and blank lines
            if len(line) == 0 or line[0] == '#':
                continue

            # parse lines so we can quickly answer containment questions
            ip, rest = line.split(maxsplit=1)
            hosts = rest.split()
            self._ips[ip].append(idx)
            for host in hosts:
                self._hosts[host].append(idx)
    
    def __contains__(self, x):
        "Check if a given host is present in this hostfile"
        return x in self._hosts
    
    def __getitem__(self, key):
        "Returns a list of lines where this host is defined."
        return self._hosts[key]
    
    def __iter__(self):
        "Returns an iterable of the hostnames in this hostfile"
        return iter(self._hosts)

    def __len__(self):
        "Return how many hosts are in this hostfile"
        return len(self._hosts)
    
    def ip_on_line(self, line: int):
        if line > len(self._contents):
            raise ValueError(f'line should be {len(self._contents)} or less')
        if line <= 0:
            raise ValueError('line must be 1 or greater')

        for ip, all_lines in self._ips.items():
            for haystack_line in all_lines:
                if haystack_line == line:
                    return ip
    
    def __format__(self, format_spec):
        """Returns a printable hostfile.
        
        format_spec can be one of:
        - raw - just return what we read, same as str()
        - clean - one line per line of the original, but cleaned up a bit (default)
        - simple - one line per unique IP address, strip comments
        """
        match format_spec:
            case 'raw':
                return str(self)
            case 'clean' | '' | None:
                return self._format_clean()
            case 'simple':
                return self._format_simple()
            case _:
                raise ValueError('format_spec not recognized')

    def _format_clean(self):
        results = []

        for line in self._contents:
            line = line.strip()
            # skip blank lines
            if len(line) == 0:
                continue
            # emit comments as-is
            if line[0] == '#':
                results.append(line)
                continue
            # otherwise this is a normal line
            ip, rest = line.split(maxsplit=1)
            hosts = rest.split()
            results.append(f"{ip}\t{' '.join(hosts)}")
        
        return '\n'.join(results)

    def _format_simple(self):
        results = ['# simplified to one line per IP']

        for ip, host_lines in self._ips.items():
            hosts = []
            for host_line in host_lines:
                _, raw_hosts = self._contents[host_line-1].strip().split()
                hosts.extend(raw_hosts.split())
            results.append(f"{ip}\t{' '.join(hosts)}")
        
        return '\n'.join(results)

    def __str__(self):
        return ''.join(self._contents)
    
    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self._contents)} lines, {len(self._ips)} unique IPs, {len(self._hosts)} unique hosts>"


if __name__ == '__main__':
    hosts = """##
# Host Database
#
# localhost is used to configure the loopback interface
# when the system is booting.  Do not change this entry.
##
127.0.0.1	localhost
255.255.255.255	broadcasthost
::1             localhost
127.0.0.1       trivia.localhost
"""
    hf = Hostfile(hosts)

    # with open('/etc/hosts') as f:
    #     hf = Hostfile(f)

    print(hf)
    print(repr(hf))
    print(format(hf, 'clean'))
    print()
    print(format(hf, 'simple'))

    print(f"{hf['localhost']=}")
    print(f"{hf['trivia.localhost']=}")

    for h in hf:
        print(h)
        for l in hf[h]:
            print('  ', hf.ip_on_line(l))
